from __future__ import annotations

import logging
from dataclasses import dataclass
from decimal import Decimal
from typing import TYPE_CHECKING

import pandas as pd

# TODO: use explicit imports until mypyc fixes attribute lookups in dataclass
# (https://github.com/mypyc/mypyc/issues/1000)
from pandas import Timestamp  # type: ignore

if TYPE_CHECKING:
    from .asset import AssetName

logger = logging.getLogger(__name__)

__all__ = ["CashTransaction"]


@dataclass(frozen=True, kw_only=True)
class Transaction:
    """A frozen record of a transaction."""

    ts: pd.Timestamp
    """Transaction time."""

    total: Decimal = Decimal(0)
    """Total transaction value.

    Negative values are costs and positive values are benefits.
    """

    desc: str = ""
    """Description."""

    def __post_init__(self):
        # since frozen we need to use obj method to cast values
        if not isinstance(self.total, Decimal):
            object.__setattr__(self, "total", Decimal(self.total))


@dataclass(frozen=True, kw_only=True)
class CashTransaction(Transaction):
    """A frozen record of a cash transaction."""


@dataclass(frozen=True, kw_only=True)
class Trade(Transaction):
    """A frozen record of a trade transaction.

    A negative `quantity` represents a sell trade and a positive
    `quantity` represents a buy trade.
    """

    quantity: Decimal
    """Traded quantity."""

    price: Decimal
    """Traded price."""

    asset_name: AssetName
    """Traded asset."""

    order_label: str | None
    """Associated Order label."""

    def __post_init__(self):
        super().__post_init__()

        if self.quantity == 0:
            raise ValueError("trade quantity cannot be zero")

        # since frozen we need to use obj method to cast values
        if not isinstance(self.price, Decimal):
            object.__setattr__(self, "price", Decimal(self.price))
        if not isinstance(self.quantity, Decimal):
            object.__setattr__(self, "quantity", Decimal(self.quantity))

        # total represents cost/benefit when buy/sell
        object.__setattr__(self, "total", -self.quantity * self.price)
        object.__setattr__(
            self,
            "desc",
            f"{'sell' if self.quantity < 0 else 'buy'} {self.asset_name}",
        )



r"""Heston Stochastic Volatility simulation.

Simulate stochastic process :math:`S_t` where,

.. math::

   dS_t = \mu S_t dt + \sqrt{\nu_t} S_t dW_t^S \\
   d\nu_t = \kappa(\theta - \nu_t) dt + \xi \sqrt{\nu_t} dW_t^{\nu}

where :math:`\mu` is the drift, :math:`\nu_t` is the variance (volatility squared)
and :math:`dW_t^S`, :math:`dW_t^{\nu_t}` are Weiner processes. :math:`\theta` is long
term variance and :math:`\kappa` is mean reversion rate. :math:`xi` is volatility of
volatility.
"""

import numpy as np

from yabte.utilities.simulation.weiner import weiner_simulate_paths


def heston_simulate_paths(
    S0: float,
    v0: float,
    mu: float,
    kappa: float,
    theta: float,
    xi: float,
    R: np.ndarray,
    T: int,
    n_steps: int,
    n_sims: int,
    rng=None,
):
    """Generate simulated paths.

    `S0` and `v0` are initial values, `mu` is the drift, `xi` is volatility of volatility, `theta` is long term variance, `R` a correlation
    matrix, `T` is the time span, `n_steps` is how many time steps,
    `n_sims` the number of simulations and `rng` a numpy random
    number generator (optional).

    TODO: support Euler–Maruyama / Milstein / Antithetic.
    """
    if rng is None:
        rng = np.random.default_rng()

    r = mu  # mu = rf in risk neutral framework
    dt = T / n_steps

    ts = np.linspace(0, T, n_steps, endpoint=False)

    # simulate 'n_sims' price paths of `k` sized asset groups with 'n_steps' timesteps
    # ws[t, sim, path]; path[0] = vol, path[1] = price
    S_ix, v_ix = 0, 1
    ws = weiner_simulate_paths(
        n_steps=n_steps, n_sims=n_sims, stdev=np.sqrt(dt), R=R, rng=rng
    )

    # calc vol
    # v[t, sim]
    v = np.zeros(shape=(n_steps, n_sims))
    v[0, :] = v0
    for i in range(1, n_steps):
        vp = v[i - 1, :]
        vp_max = np.clip(vp, 0, None)
        v[i, :] = (
            vp
            + kappa * (theta - vp_max) * ts[i]
            + xi * np.sqrt(vp_max) * ws[i, :, v_ix]
        )

    # S[t, sim]
    S = S0 * np.exp((r - v**2) * ts.reshape(n_steps, -1) + v * ws[:, :, S_ix])

    return S, v



import pandas as pd


def crossover(series1: pd.Series, series2: pd.Series) -> bool:
    """Determine if two series cross over one another. Returns `True` if `series1` just
    crosses above `series2`.

        >>> crossover(self.data.Close, self.sma)
        True
    """
    try:
        return (
            series1.iloc[-2] < series2.iloc[-2] and series1.iloc[-1] > series2.iloc[-1]
        )
    except IndexError:
        return False



import unittest

import numpy as np
import numpy.linalg as la

import yabte.utilities.pandas_extension  # noqa
from yabte.tests._helpers import generate_nasdaq_dataset
from yabte.tests._unittest_numpy_extensions import NumpyTestCase
from yabte.utilities.lagrangian import Lagrangian


class UtilitiesTestCase(NumpyTestCase):
    @classmethod
    def setUpClass(cls):
        cls.assets, cls.df_combined = generate_nasdaq_dataset()
        cls.closes = cls.df_combined.loc[:, (slice(None), "Close")].droplevel(
            axis=1, level=1
        )
        cls.returns = cls.closes.prc.log_returns

    def test_lagrangian(self):
        Sigma = self.returns.cov()
        mu = self.closes.prc.capm_returns()
        r = 0.1

        # solve algebraically
        m = len(mu)
        ones = np.ones(m)
        SigmaInv = la.inv(Sigma)
        A = mu.T @ SigmaInv @ ones
        B = mu.T @ SigmaInv @ mu
        C = ones.T @ SigmaInv @ ones
        D = B * C - A * A
        l1 = (C * r - A) / D
        l2 = (B - A * r) / D
        w = SigmaInv @ (l1 * mu + l2 * ones)

        # sanity checks
        self.numpyAssertAllclose(w.sum(), 1)
        self.numpyAssertAllclose(w @ mu, r)

        # test numerical
        L = Lagrangian(
            objective=lambda x: x.T @ Sigma @ x / 2,
            constraints=[
                lambda x: r - x.T @ mu,
                lambda x: 1 - x.T @ ones,
            ],
            x0=np.ones(m) / m,
        )
        wn = L.fit()

        self.numpyAssertAllclose(wn, w)


if __name__ == "__main__":
    unittest.main()



r"""Weiner process.

Weiner process :math:`W_t` such that,

#. :math:`W_0 = 0`
#. :math:`\forall t>0, u \ge  0, s<t \Rightarrow W_{t+u}-W_{t} \text{ independent of } W_s`
#. :math:`W_{t+u}-W_{t}\sim {\mathcal {N}}(0,u)`
#. :math:`W_t` is continuous
"""

import numpy as np


def weiner_simulate_paths(
    n_steps: int,
    n_sims: int = 1,
    stdev: float | np.ndarray = 1,
    R: np.ndarray = np.array([[1]]),
    rng=None,
):
    """Generate simulated Weiner paths.

    `stdev` is the increment size, `R` a correlation matrix, `n_steps`
    is how many time steps, `n_sims` the number of simulations and `rng`
    a numpy random number generator (optional). If `stdev` is a scalar
    it will be broadcasted to the size of `n_sims`.
    """

    R = np.atleast_2d(R)
    stdev = np.resize(stdev, n_sims).reshape(n_sims, 1)

    if rng is None:
        rng = np.random.default_rng()

    k = R.shape[0]

    # simulate 'n_sims' price paths of `k` sized asset groups with 'n_steps' timesteps
    dws = rng.multivariate_normal(
        mean=np.zeros(k), cov=R, size=(n_steps - 1, n_sims), check_valid="raise"
    )

    # use cumsum as speed up
    ws = np.concatenate([np.zeros((1, n_sims, k)), np.cumsum(dws, axis=0)])

    return stdev * ws



import logging
import unittest
from decimal import Decimal

import numpy as np
import pandas as pd

import yabte.utilities.pandas_extension
from yabte.backtest import (
    BasketOrder,
    Book,
    OHLCAsset,
    OrderSizeType,
    OrderStatus,
    PositionalBasketOrder,
    PositionalOrder,
    SimpleOrder,
    Strategy,
    StrategyRunner,
)
from yabte.tests._helpers import generate_nasdaq_dataset
from yabte.utilities.strategy_helpers import crossover

logger = logging.getLogger(__name__)


class TestSMAXOStrat(Strategy):
    def init(self):
        p = self.params
        days_short = p.get("days_short", 10)
        days_long = p.get("days_long", 20)

        close_sma_short = (
            self.data.loc[:, (slice(None), "Close")]
            .rolling(days_short)
            .mean()
            .rename({"Close": "CloseSMAShort"}, axis=1, level=1)
        )
        close_sma_long = (
            self.data.loc[:, (slice(None), "Close")]
            .rolling(days_long)
            .mean()
            .rename({"Close": "CloseSMALong"}, axis=1, level=1)
        )
        self.data = pd.concat(
            [self.data, close_sma_short, close_sma_long], axis=1
        ).sort_index(axis=1)

    def on_close(self):
        p = self.params
        symbol = p.get("symbol", "GOOG")

        df = self.data[symbol]
        ix_2d = df.index[-2:]
        data = df.loc[ix_2d, ("CloseSMAShort", "CloseSMALong")].dropna()
        if len(data) == 2:
            if crossover(data.CloseSMAShort, data.CloseSMALong):
                self.orders.append(SimpleOrder(asset_name=symbol, size=100))
            elif crossover(data.CloseSMALong, data.CloseSMAShort):
                self.orders.append(SimpleOrder(asset_name=symbol, size=-100))


class TestSMAXOMultipleBookStrat(TestSMAXOStrat):
    def on_close(self):
        # create some orders

        for symbol in ["GOOG", "MSFT"]:
            book_name = f"{symbol}_BOOK"
            df = self.data[symbol]
            ix_2d = df.index[-2:]
            data = df.loc[ix_2d, ("CloseSMAShort", "CloseSMALong")].dropna()
            if len(data) == 2:
                if crossover(data.CloseSMAShort, data.CloseSMALong):
                    self.orders.append(
                        SimpleOrder(book=book_name, asset_name=symbol, size=-100)
                    )
                elif crossover(data.CloseSMALong, data.CloseSMAShort):
                    self.orders.append(
                        SimpleOrder(book=book_name, asset_name=symbol, size=100)
                    )


class TestPosOrderSizeStrat(Strategy):
    def on_close(self):
        p = self.params
        size_type = p.size_type
        size_factor = p.size_factor
        symbol = p.get("symbol", "GOOG")

        ix = self.data.index.get_loc(self.ts)
        if ix in [100, 201, 300, 401]:
            quantity = size_factor * (-1) ** ix
            self.orders.append(
                PositionalOrder(asset_name=symbol, size=quantity, size_type=size_type)
            )
        elif ix in [1000]:
            self.orders.append(PositionalOrder(asset_name=symbol, size=0))


class TestSpreadSimpleStrat(Strategy):
    def init(self):
        p = self.params
        s = self.data[p.s1].Close - p.factor * self.data[p.s2].Close
        self.data.loc[:, ("SPREAD", "Close")] = s
        self.mu = s.mean()
        self.sigma = s.std()

    def on_close(self):
        p = self.params
        s = self.data["SPREAD"].Close.iloc[-1]
        if s < self.mu - 0.5 * self.sigma:
            self.orders.append(PositionalOrder(asset_name=p.s1, size=100))
            self.orders.append(PositionalOrder(asset_name=p.s2, size=p.factor * 100))
        elif s > self.mu + 0.5 * self.sigma:
            self.orders.append(PositionalOrder(asset_name=p.s1, size=-100))
            self.orders.append(PositionalOrder(asset_name=p.s2, size=-p.factor * 100))
        elif abs(s) < 0.1 * self.sigma:
            self.orders.append(PositionalOrder(asset_name=p.s1, size=0))
            self.orders.append(PositionalOrder(asset_name=p.s2, size=0))


class TestBasketOrderSizeStrat(Strategy):
    def on_close(self):
        p = self.params
        size_type = p.size_type
        symbols = ["AAPL", "AMZN", "GOOG", "META"]
        weights = [1, 2, 3, 4]

        ix = self.data.index.get_loc(self.ts)
        if ix in [100, 201, 300, 401]:
            self.orders.append(
                BasketOrder(
                    asset_names=symbols, weights=weights, size=1, size_type=size_type
                )
            )
        elif ix in [1000]:
            self.orders.append(
                PositionalBasketOrder(
                    asset_names=symbols, weights=[0] * len(symbols), size=1
                )
            )


class StrategyRunnerTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.assets, cls.df_combined = generate_nasdaq_dataset()

    def test_sma_crossover(self):
        sr = StrategyRunner(
            data=self.df_combined,
            assets=self.assets,
            strategies=[TestSMAXOStrat()],
        )
        srr = sr.run()

        th = srr.transaction_history
        th["nc"] = -th.quantity * th.price
        bch = (
            th.pivot_table(index="ts", columns="book", values="nc", aggfunc="sum")
            .cumsum()
            .reindex(sr.data.index)
            .ffill()
            .fillna(0)
        )
        self.assertTrue(
            np.all(
                np.isclose(
                    bch.astype("float64"),
                    srr.book_history.loc[:, (slice(None), "cash")].droplevel(
                        axis=1, level=1
                    ),
                )
            )
        )

    def test_multiple_books(self):
        books = [
            Book(name="MSFT_BOOK", cash=Decimal("1000000")),
            Book(name="GOOG_BOOK", cash=Decimal("1000000")),
        ]

        sr = StrategyRunner(
            data=self.df_combined,
            assets=self.assets,
            strategies=[TestSMAXOMultipleBookStrat()],
            books=books,
        )
        srr = sr.run()

        th = srr.transaction_history
        self.assertEqual(len(th.book.unique()), 2)
        bh = srr.book_history
        self.assertEqual(len(bh.columns.levels[0]), 2)

    def test_positional_orders_quantity(self):
        # test using quantities
        sr = StrategyRunner(
            data=self.df_combined,
            assets=self.assets,
            strategies=[TestPosOrderSizeStrat()],
        )
        srr = sr.run(params={"size_type": OrderSizeType.QUANTITY, "size_factor": 100})

        # 8 = ococococ
        self.assertEqual(len(srr.books[0].transactions), 8)

    def test_positional_orders_notional(self):
        # test using notionals
        sr = StrategyRunner(
            data=self.df_combined,
            assets=self.assets,
            strategies=[TestPosOrderSizeStrat()],
        )
        srr = sr.run(params={"size_type": OrderSizeType.NOTIONAL, "size_factor": 1000})

        # 8 = ococococ
        self.assertEqual(len(srr.books[0].transactions), 8)

    def test_positional_orders_book_percent(self):
        # test using book percent

        # books default to zero cash
        book = Book(name="Main", cash=Decimal("1000000"))

        sr = StrategyRunner(
            data=self.df_combined,
            assets=self.assets,
            strategies=[TestPosOrderSizeStrat()],
            books=[book],
        )
        srr = sr.run(
            params={
                "size_type": OrderSizeType.BOOK_PERCENT,
                "size_factor": 10000,
            }
        )

        # 8 = ococococ
        self.assertEqual(len(srr.books[0].transactions), 8)

    def test_spread_simple(self):
        params = {
            "s1": "GOOG",
            "s2": "MSFT",
            "factor": 4.5,
        }
        sr = StrategyRunner(
            data=self.df_combined,
            assets=self.assets,
            strategies=[TestSpreadSimpleStrat()],
        )
        srr = sr.run(params)

        df_trades = pd.DataFrame(srr.books[0].transactions)
        self.assertEqual(len(df_trades), 6)
        self.assertEqual(len(df_trades.query("asset_name == 'GOOG'")), 3)
        self.assertEqual(len(df_trades.query("asset_name == 'MSFT'")), 3)

    def test_basket_order_quantity(self):
        # test using quantities
        sr = StrategyRunner(
            data=self.df_combined,
            assets=self.assets,
            strategies=[TestBasketOrderSizeStrat()],
        )
        srr = sr.run(params={"size_type": OrderSizeType.QUANTITY})

        # 20 = 4 x ttttc
        self.assertEqual(len(srr.books[0].transactions), 20)

    def test_on_open_masking(self):
        class TestOnOpenMaskStrat(Strategy):
            def on_open(self2):
                # only latest record masked
                if len(self2.data) > 1:
                    self.assertTrue(
                        all(
                            self2.data.iloc[:-1]
                            .loc[
                                :,
                                (
                                    slice(None),
                                    ("Open", "High", "Low", "Close", "Volume"),
                                ),
                            ]
                            .notnull()
                        )
                    )

                # these fields should be masked at open
                self.assertTrue(
                    all(
                        self2.data.iloc[-1:]
                        .loc[:, (slice(None), ("High", "Low", "Close", "Volume"))]
                        .isnull()
                    )
                )

                # this field should be unmasked
                self.assertTrue(
                    all(self2.data.iloc[-1:].loc[:, (slice(None), "Open")].notnull())
                )

            def on_close(self2):
                # at close all fields available
                self.assertTrue(
                    all(
                        self2.data.loc[
                            :, (slice(None), ("Open", "High", "Low", "Close", "Volume"))
                        ].notnull()
                    )
                )

        data = pd.DataFrame(
            [
                [100] * 10,
                [100] * 10,
                [100] * 10,
            ],
            columns=pd.MultiIndex.from_product(
                [["ACME", "BOKO"], ["High", "Low", "Open", "Close", "Volume"]]
            ),
            index=pd.date_range(start="20180102", periods=3, freq="B"),
        )

        sr = StrategyRunner(
            data=data,
            assets=[OHLCAsset(name="ACME"), OHLCAsset(name="BOKO")],
            strategies=[TestOnOpenMaskStrat()],
        )
        sr.run()

    def test_limit_order(self):
        class LimitOrder(SimpleOrder):
            def pre_execute_check(self, ts, tp):
                # if goes above 110 then cancel
                if tp > 110:
                    return OrderStatus.CANCELLED
                # if drops below 90 then complete order
                elif tp < 90:
                    return None
                # otherwise leave open for another day
                return OrderStatus.OPEN

        class TestLimitOrderStrat(Strategy):
            def on_close(self):
                ix = self.data.index.get_loc(self.ts)
                if ix == 0:
                    self.orders.append(LimitOrder(asset_name="ACME", size=100))

        for ix, (data_arr, op_status, ou_status) in enumerate(
            [
                (
                    [
                        [105],
                        [115],
                        [110],
                    ],
                    [OrderStatus.CANCELLED],
                    [],
                ),
                (
                    [
                        [95],
                        [100],
                        [105],
                    ],
                    [],
                    [OrderStatus.OPEN],
                ),
                (
                    [
                        [95],
                        [100],
                        [85],
                    ],
                    [OrderStatus.COMPLETE],
                    [],
                ),
            ]
        ):
            with self.subTest(i=ix):
                data = pd.DataFrame(
                    data_arr,
                    columns=pd.MultiIndex.from_product([["ACME"], ["Close"]]),
                    index=pd.date_range(
                        start="20180102", periods=len(data_arr), freq="B"
                    ),
                )

                sr = StrategyRunner(
                    data=data,
                    assets=[OHLCAsset(name="ACME", denom="USD")],
                    strategies=[TestLimitOrderStrat()],
                )
                srr = sr.run()

                self.assertListEqual(
                    op_status, [o.status for o in srr.orders_processed]
                )
                self.assertListEqual(
                    ou_status, [o.status for o in srr.orders_unprocessed]
                )

    def test_stop_loss_order(self):
        class StopLossOrder(SimpleOrder):
            def pre_execute_check(self, ts, tp):
                # if drops below 90 then complete stop order
                if tp < 90:
                    return None
                # otherwise leave open for another day
                return OrderStatus.OPEN

        class OrderWithStopLosses(SimpleOrder):
            def post_complete(self, trades):
                self.suborders.extend(
                    [
                        StopLossOrder(
                            asset_name=t.asset_name,
                            size=-t.quantity,
                            label="my_stop",
                        )
                        for t in trades
                    ]
                )

        class TestStopLossOrderStrat(Strategy):
            def on_close(self):
                ix = self.data.index.get_loc(self.ts)
                if ix == 0:
                    self.orders.append(OrderWithStopLosses(asset_name="ACME", size=100))

        for ix, (data_arr, op_status, ou_status) in enumerate(
            [
                (
                    [
                        [105],
                        [115],
                        [110],
                    ],
                    [(OrderStatus.COMPLETE, None)],
                    [(OrderStatus.OPEN, "my_stop")],
                ),
                (
                    [
                        [95],
                        [100],
                        [85],
                    ],
                    [(OrderStatus.COMPLETE, None), (OrderStatus.COMPLETE, "my_stop")],
                    [],
                ),
            ]
        ):
            with self.subTest(i=ix):
                data = pd.DataFrame(
                    data_arr,
                    columns=pd.MultiIndex.from_product([["ACME"], ["Close"]]),
                    index=pd.date_range(
                        start="20180102", periods=len(data_arr), freq="B"
                    ),
                )

                sr = StrategyRunner(
                    data=data,
                    assets=[OHLCAsset(name="ACME", denom="USD")],
                    strategies=[TestStopLossOrderStrat()],
                )
                srr = sr.run()

                self.assertListEqual(
                    op_status, [(o.status, o.label) for o in srr.orders_processed]
                )
                self.assertListEqual(
                    ou_status, [(o.status, o.label) for o in srr.orders_unprocessed]
                )

    def test_priority(self):
        class TestPriorityStrat(Strategy):
            def on_close(self):
                ix = self.data.index.get_loc(self.ts)
                if ix == 0:
                    self.orders.append(
                        SimpleOrder(asset_name="ACME", size=100, priority=2)
                    )
                    self.orders.append(
                        SimpleOrder(asset_name="ACME", size=200, priority=3)
                    )
                    self.orders.append(
                        SimpleOrder(asset_name="ACME", size=300, priority=1)
                    )

        data_arr = [
            [105],
            [115],
            [110],
        ]

        data = pd.DataFrame(
            data_arr,
            columns=pd.MultiIndex.from_product([["ACME"], ["Close"]]),
            index=pd.date_range(start="20180102", periods=len(data_arr), freq="B"),
        )

        sr = StrategyRunner(
            data=data,
            assets=[OHLCAsset(name="ACME", denom="USD")],
            strategies=[TestPriorityStrat()],
        )
        srr = sr.run()

        self.assertListEqual(
            srr.transaction_history.quantity.to_list(),
            [Decimal("200.00"), Decimal("100.00"), Decimal("300.00")],
        )

    def test_order_key(self):
        class LimitOrder(SimpleOrder):
            # this limit won't be met in test
            def pre_execute_check(self, ts, tp):
                # if goes above 200 then complete order
                if tp > 200:
                    return None
                # otherwise leave open for another day
                return OrderStatus.OPEN

        class TestOrderkeyStrat(Strategy):
            def on_close(self):
                ix = self.data.index.get_loc(self.ts)
                if ix == 0:
                    self.orders.append(
                        LimitOrder(asset_name="ACME", size=100, key="my_key")
                    )
                    self.orders.append(LimitOrder(asset_name="ACME", size=200))
                elif ix == 1:
                    self.orders.append(
                        LimitOrder(asset_name="ACME", size=300, key="my_key")
                    )

        data_arr = [
            [105],
            [115],
            [110],
        ]

        data = pd.DataFrame(
            data_arr,
            columns=pd.MultiIndex.from_product([["ACME"], ["Close"]]),
            index=pd.date_range(start="20180102", periods=len(data_arr), freq="B"),
        )

        sr = StrategyRunner(
            data=data,
            assets=[OHLCAsset(name="ACME", denom="USD")],
            strategies=[TestOrderkeyStrat()],
        )
        srr = sr.run()

        self.assertListEqual(
            [
                (OrderStatus.REPLACED, "my_key", Decimal("100")),
            ],
            [(o.status, o.key, o.size) for o in srr.orders_processed],
        )

        self.assertListEqual(
            [
                (OrderStatus.OPEN, None, Decimal("200")),
                (OrderStatus.OPEN, "my_key", Decimal("300")),
            ],
            [(o.status, o.key, o.size) for o in srr.orders_unprocessed],
        )

    def test_run_batch(self):
        book = Book(name="Main", cash=Decimal("100000"))

        sr = StrategyRunner(
            data=self.df_combined,
            assets=self.assets,
            strategies=[TestSMAXOStrat()],
            books=[book],
        )

        param_iter = [
            {"days_long": n, "days_short": m}
            for n, m in zip([20, 30, 40, 50], [5, 10, 15, 20])
            if n > m
        ]

        srrs = sr.run_batch(param_iter)

        self.assertEqual(len(srrs), len(param_iter))

        # check we have distinct sharpe ratios for each param set
        sharpes = {
            srr.book_history.loc[:, ("Main", "total")].prc.sharpe_ratio()
            for srr in srrs
        }
        self.assertEqual(len(sharpes), len(param_iter))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import numpy.linalg as la

import yabte.utilities.pandas_extension  # noqa
from yabte.tests._helpers import generate_nasdaq_dataset
from yabte.tests._unittest_numpy_extensions import NumpyTestCase
from yabte.utilities.lagrangian import Lagrangian
from yabte.utilities.portopt.hierarchical_risk_parity import hrp
from yabte.utilities.portopt.inverse_volatility import inverse_volatility
from yabte.utilities.portopt.minimum_variance import (
    minimum_variance,
    minimum_variance_numeric,
    minimum_variance_numeric_slsqp,
)


class PortOptTestCase(NumpyTestCase):
    @classmethod
    def setUpClass(cls):
        cls.assets, cls.df_combined = generate_nasdaq_dataset()
        cls.closes = cls.df_combined.loc[:, (slice(None), "Close")].droplevel(
            axis=1, level=1
        )
        cls.returns = cls.closes.prc.log_returns

    def test_min_var(self):
        Sigma = self.returns.cov()
        mu = self.closes.prc.capm_returns()
        r = 0.1

        w = minimum_variance(Sigma, mu, r)
        wn = minimum_variance_numeric(Sigma, mu, r)
        wn2 = minimum_variance_numeric_slsqp(Sigma, mu, r)

        self.numpyAssertAllclose(w, wn)
        # slsqp not as numerically accurate as other two methods
        self.numpyAssertAllclose(w, wn2, rtol=1e-06)

    def test_hrp(self):
        R = self.returns.corr()
        sigma = self.returns.std()

        w = hrp(R, sigma)

        self.numpyAssertAllclose(w.sum(), 1)

    def test_ivp(self):
        Sigma = self.returns.cov()

        w = inverse_volatility(Sigma)

        self.numpyAssertAllclose(w.sum(), 1)


if __name__ == "__main__":
    unittest.main()



import concurrent.futures
import logging
from concurrent.futures import ProcessPoolExecutor
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional

import pandas as pd

# TODO: use explicit imports until mypyc fixes attribute lookups in dataclass
# (https://github.com/mypyc/mypyc/issues/1000)
from pandas import DataFrame, Series, Timestamp  # type: ignore

from .asset import Asset, AssetName
from .book import Book, BookMandate, BookName
from .order import Order, Orders, OrderStatus
from .strategy import Strategy

logger = logging.getLogger(__name__)

__all__ = ["StrategyRunner"]


def _check_data(df, asset_map):
    """Check data structure correct."""

    if not isinstance(df.index, pd.DatetimeIndex):
        raise ValueError("data index must be a datetimeindex")
    if not df.index.is_monotonic_increasing:
        raise ValueError("data needs to have increasing index")
    if not df.index.is_unique:
        raise ValueError("data index must be unique")

    # column level 1 = asset, level 2 = field
    if not isinstance(df.columns, pd.MultiIndex):
        raise ValueError("data columns must be multindex asset/field")
    if len(df.columns.levels) != 2:
        raise ValueError("data columns multiindex must have 2 levels")

    # for cartesian products
    data_labels_data = set(df.columns.levels[0])
    data_labels_asset = {a.data_label for a in asset_map.values()}
    assets_missing_data = data_labels_asset - data_labels_data
    if len(assets_missing_data):
        raise ValueError(
            f"some assets are missing corresponding data: {assets_missing_data}"
        )

    # check and fix data for each asset
    dfs = {
        asset.data_label: asset.check_and_fix_data(asset._filter_data(df))
        for asset_name, asset in asset_map.items()
    }

    return pd.concat(dfs, axis=1)


@dataclass(kw_only=True)
class StrategyRunnerResult:
    books: List[Book]
    """Books used within this result."""

    strategies: List[Strategy]
    """Strategies used within this result."""

    assets: List[Asset]
    """Assets used within this result."""

    _orders_unprocessed: Orders = field(default_factory=Orders)

    @property
    def orders_unprocessed(self) -> Orders:
        """Unprocessed orders queue."""
        return self._orders_unprocessed

    _orders_processed: List[Order] = field(default_factory=list)

    @property
    def orders_processed(self) -> List[Order]:
        """Processed orders list."""
        return self._orders_processed

    _book_history: Optional[pd.DataFrame] = None

    @property
    def book_history(self) -> pd.DataFrame:
        """Dataframe with book cash, mtm and total value history."""
        return pd.concat({b.name: b.history for b in self.books}, axis=1)

    @property
    def transaction_history(self) -> pd.DataFrame:
        """Dataframe with trade history."""
        return pd.concat(
            [pd.DataFrame(bk.transactions).assign(book=bk.name) for bk in self.books]
        )


@dataclass(kw_only=True)
class StrategyRunner:
    """Encapsulates the execution of multiple strategies.

    Orders are captured in `orders_processed` and `orders_unprocessed`.
    `books` is a list of books and if none provided a single book is
    created called 'Main'. After execution summary book and trade
    histories are captured in `book_history` and `transaction_history`.
    """

    data: pd.DataFrame = field()
    """Dataframe of price data including columns High, Low, Open, Close, Volume for each
    asset.

    Both asset name and field make a multiindex column. The index should consist of
    order pandas timestamps.
    """

    assets: List[Asset]
    """Assets available to strategy."""

    strategies: List[Strategy]
    """Strategies to be called within this runner."""

    mandates: Dict[AssetName, BookMandate] = field(default_factory=dict)
    """Dictionary of asset mandates (experimental)."""

    books: List[Book] = field(default_factory=list)
    """Books available to strategies.

    If not supplied will be populated with single book named 'Main' denominated in USD.
    """

    def __post_init__(self):
        asset_map = {asset.name: asset for asset in self.assets}
        self.data = _check_data(self.data, asset_map)

        # set up books
        if not self.books:
            self.books = [Book(name="Main", mandates=self.mandates)]

    def run(self, params: Dict[str, Any] = None) -> StrategyRunnerResult:
        """Execute each strategy through time."""

        params = pd.Series(params or {}, dtype=object)

        srr = StrategyRunnerResult(
            assets=deepcopy(self.assets),  # should not be mutable
            books=deepcopy(self.books),
            strategies=deepcopy(self.strategies),
        )

        # made available where necessary
        asset_map = {asset.name: asset for asset in srr.assets}
        book_map = {book.name: book for book in srr.books}

        # calendar
        calendar = self.data.index

        for strat in srr.strategies:
            strat.params = params
            strat.orders = srr._orders_unprocessed
            strat.books = book_map
            strat.assets = asset_map

            strat._data_lock = False
            strat.data = deepcopy(self.data)
            strat.init()
            strat._data_lock = True

        # run event loop
        for ts in calendar:
            logger.info(f"Processing timestep {ts}")

            # open
            for strat in srr.strategies:
                # provide window
                strat._set_ts(ts)
                strat._mask_open = True
                strat.on_open()
                strat._mask_open = False

            # order applied with ts's data
            day_data = self.data.loc[ts, :]

            # sort orders by priority
            srr._orders_unprocessed.sort_by_priority()

            # process orders
            orders_next_ts = []
            while srr._orders_unprocessed:
                order = srr._orders_unprocessed.popleft()

                # set book attribute if needed
                if not isinstance(order.book, Book):
                    # fall back to first available book
                    order.book = book_map.get(order.book, srr.books[0])

                order.apply(ts, day_data, asset_map)

                # add any child orders to next ts
                orders_next_ts.extend(order.suborders)

                if order.status == OrderStatus.OPEN:
                    orders_next_ts.append(order)
                else:
                    srr.orders_processed.append(order)

            # extend with orders for next ts
            srr._orders_unprocessed.extend(orders_next_ts)

            # remove older duplicate orders
            replaced = srr._orders_unprocessed.remove_duplicate_keys()
            srr.orders_processed.extend(replaced)

            # close
            for strat in srr.strategies:
                # provide window
                strat._set_ts(ts)
                strat.on_close()

            # run book end-of-day tasks
            for book in srr.books:
                book.eod_tasks(ts, day_data, asset_map)

        return srr

    def run_batch(
        self,
        params_iterable: Iterable[Dict[str, Any]],
        executor: ProcessPoolExecutor | None = None,
    ) -> List[StrategyRunnerResult]:
        """Run a set of parameter combinations."""

        executor = executor or concurrent.futures.ThreadPoolExecutor()
        with executor:
            return list(executor.map(self.run, params_iterable))



import unittest

import numpy as np

from yabte.tests._unittest_numpy_extensions import NumpyTestCase
from yabte.utilities.simulation.geometric_brownian_motion import gbm_simulate_paths
from yabte.utilities.simulation.heston import heston_simulate_paths
from yabte.utilities.simulation.weiner import weiner_simulate_paths

HAS_SCIPY = True
try:
    import scipy.stats as stats
except:
    HAS_SCIPY = False


def _gbm_recover_weiner(T, N, M, r, p, sigma):
    # duplicate copies of time axis to simplify broadcasting later
    ts = np.linspace(0, T, N, endpoint=False)
    ts = np.repeat(ts[:, np.newaxis], M, axis=1)[:, :, np.newaxis]
    dt = T / N
    r_ = np.atleast_1d(r)
    sigma_ = np.atleast_1d(sigma)
    ws = (np.log(p) - (r_ - sigma_**2 / 2) * ts) / sigma_ / np.sqrt(dt)
    return ws


class SimulationTestCase(NumpyTestCase):
    def setUp(self):
        self.rng = np.random.default_rng(12345)
        super().setUp()

    def test_weiner_simple(self):
        R = [[1, 0.9], [0.9, 1]]
        N = 101
        M = 3

        # simulate data
        # p[steps, sims, path]
        ws = weiner_simulate_paths(n_steps=N, n_sims=M, stdev=1, R=R, rng=self.rng)
        dws = np.diff(ws, axis=0)

        # test weiner properities
        self.assertTrue((ws[0] == 0).all())

        # check correlation (TODO: tolerance a bit poor here)
        for m in range(M):
            R_ = np.corrcoef(dws[:, m, 0], dws[:, m, 1])
            self.numpyAssertAllclose(R, R_, atol=0.03)

    @unittest.skipUnless(HAS_SCIPY, "needs scipy")
    def test_weiner_simple_ks(self):
        R = [[1, 0.9], [0.9, 1]]
        N = 101
        M = 3
        K = len(R)

        # simulate data
        # p[steps, sims, path]
        ws = weiner_simulate_paths(n_steps=N, n_sims=M, stdev=1, R=R, rng=self.rng)
        dws = np.diff(ws, axis=0)

        # normal differences
        for m, k in zip(range(M), range(K)):
            ksr = stats.kstest(dws[:, m, k], "norm", args=(0, 1))
            self.assertGreater(ksr.pvalue, 0.01)

    def test_gbm_simple(self):
        r = [0.05, 0.01]
        R = [[1, 0.9], [0.9, 1]]
        sigma = 0.2
        N = 101
        T = N / 365
        M = 3
        K = len(r)

        # simulate data
        # p[steps, sims, path]
        p = gbm_simulate_paths(
            S0=1, mu=r, sigma=sigma, R=R, T=T, n_steps=N, n_sims=M, rng=self.rng
        )

        # recover weiner process
        ws = _gbm_recover_weiner(T, N, M, r, p, sigma)
        dws = np.diff(ws, axis=0)

        # test weiner properities
        self.assertTrue((ws[0] == 0).all())

        # check correlation
        for m in range(M):
            R_ = np.corrcoef(dws[:, m, 0], dws[:, m, 1])
            self.numpyAssertAllclose(R, R_, atol=0.02)

    @unittest.skipUnless(HAS_SCIPY, "needs scipy")
    def test_gbm_simple_ks(self):
        r = [0.05, 0.01]
        R = [[1, 0.9], [0.9, 1]]
        sigma = 0.2
        N = 101
        T = N / 365
        M = 3
        K = len(r)

        # simulate data
        # p[steps, sims, path]
        p = gbm_simulate_paths(
            S0=1, mu=r, sigma=sigma, R=R, T=T, n_steps=N, n_sims=M, rng=self.rng
        )

        # recover weiner process
        ws = _gbm_recover_weiner(T, N, M, r, p, sigma)
        dws = np.diff(ws, axis=0)

        # normal differences
        for m, k in zip(range(M), range(K)):
            ksr = stats.kstest(dws[:, m, k], "norm", args=(0, 1))
            self.assertGreater(ksr.pvalue, 0.01)

    def test_heston_smoke(self):
        kappa = 4
        theta = 0.02
        v0 = 0.02
        sigma = 0.9
        r = 0.02
        S0 = 100
        T = 1
        R = 0.9

        S, v = heston_simulate_paths(
            S0=S0,
            v0=v0,
            mu=r,
            kappa=kappa,
            theta=theta,
            xi=sigma,
            R=np.array([[1, R], [R, 1]]),
            T=T,
            n_steps=100,
            n_sims=1,
        )

        # TODO: better tests
        self.assertEqual(sum(S.shape), 101)
        self.assertEqual(sum(v.shape), 101)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import logging
from collections import Counter, deque
from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum
from typing import Dict, Iterable, List, Optional, Tuple

import pandas as pd
from mypy_extensions import mypyc_attr

from ._helpers import ensure_decimal, ensure_enum
from .asset import Asset, AssetName
from .book import Book, BookName
from .transaction import Trade

logger = logging.getLogger(__name__)

__all__ = ["SimpleOrder", "PositionalOrder", "BasketOrder", "PositionalBasketOrder"]


class OrderStatus(Enum):
    """Various statuses."""

    MANDATE_FAILED = 1
    """Order failed due to mandate."""

    CANCELLED = 2
    """Order was cancelled."""

    OPEN = 3
    """Order is open."""

    COMPLETE = 4
    """Order completed succesfully."""

    REPLACED = 5
    """Order was replaced."""


class OrderSizeType(Enum):
    """Various size types."""

    QUANTITY = 1
    """Size is a quantity."""

    NOTIONAL = 2
    """Size represent notional amount."""

    BOOK_PERCENT = 3
    """Size is a percentage of book value."""


@mypyc_attr(allow_interpreted_subclasses=True)
@dataclass(kw_only=True)
class Order:
    """Base class for all orders."""

    status: OrderStatus = OrderStatus.OPEN
    """Status of order."""

    book: BookName | Book | None = field(repr=False, default=None)
    """Target book."""

    suborders: List[Order] = field(default_factory=list)
    """Additional orders to be executed the following timestep."""

    label: Optional[str] = None
    """Label to assist in matching / filtering."""

    priority: int = 0
    """Each day orders are sorted by this field and executed in order."""

    key: Optional[str] = None
    """Unique key for this order.

    If a key is set then only the newest order with this key is kept. Older orders with
    the same key will be removed.
    """

    def __post_init__(self):
        pass

    def _book_trades(self, trades):
        # test then book trades, do any post complete tasks
        if self.book.test_trades(trades):
            self.book.add_transactions(trades)
            self.status = OrderStatus.COMPLETE
            self.post_complete(trades)

        else:
            self.status = OrderStatus.MANDATE_FAILED

    def post_complete(self, trades: List[Trade]):
        """Called after and with trades that have been successfully booked.

        It can append new orders to suborders for execution in the following timestep.
        """
        pass

    def apply(self, ts: pd.Timestamp, day_data: pd.Series, asset_map: Dict[str, Asset]):
        """Applies order to `self.book` for time `ts` using provided `day_data` and
        dictionary of asset information `asset_map`."""
        raise NotImplementedError("The apply methods needs to be implemented.")


class Orders:
    """Double ended queue of orders."""

    def __init__(self):
        self.deque = deque()

    def __len__(self):
        return len(self.deque)

    def __iter__(self):
        return iter(self.deque)

    def popleft(self):
        return self.deque.popleft()

    def append(self, order: Order):
        return self.deque.append(order)

    def extend(self, orders: Iterable[Order]):
        return self.deque.extend(orders)

    def sort_by_priority(self):
        """Sorts orders by order priority."""
        ou_sorted = sorted(self.deque, key=lambda o: o.priority, reverse=True)
        self.deque.clear()
        self.deque.extend(ou_sorted)

    def remove_duplicate_keys(self) -> List[Order]:
        """Remove older orders with same key.

        Returns a list of orders than were removed with status set to REPLACED.
        """
        removed = []
        cntr = Counter(o.key for o in self.deque if o.key is not None)
        if any(v > 1 for v in cntr.values()):
            kept = []
            while self.deque:
                o = self.deque.popleft()
                if o.key in cntr and cntr[o.key] > 1:
                    o.status = OrderStatus.REPLACED
                    removed.append(o)
                    cntr[o.key] -= 1
                else:
                    kept.append(o)
            self.deque.clear()
            self.deque.extend(kept)

        return removed


@mypyc_attr(allow_interpreted_subclasses=True)
@dataclass(kw_only=True)
class SimpleOrder(Order):
    """Simple market order."""

    asset_name: AssetName
    """Asset name for order."""

    size: Decimal
    """Order size."""

    size_type: OrderSizeType = OrderSizeType.QUANTITY
    """Order size type.

    Can be a quantity, notional or book percent.
    """

    def __post_init__(self):
        super().__post_init__()
        self.size = ensure_decimal(self.size)
        self.size_type = ensure_enum(self.size_type, OrderSizeType)

    def _calc_quantity_price(
        self, day_data: pd.Series, asset_map: Dict[str, Asset]
    ) -> Tuple[Decimal, Decimal]:
        asset = asset_map[self.asset_name]
        asset_day_data = asset._filter_data(day_data)
        trade_price = asset.intraday_traded_price(asset_day_data, size=self.size)

        if self.size_type == OrderSizeType.QUANTITY:
            quantity = self.size
        elif self.size_type == OrderSizeType.NOTIONAL:
            quantity = self.size / trade_price
        elif self.size_type == OrderSizeType.BOOK_PERCENT:
            assert isinstance(self.book, Book)  # to please mypy
            quantity = self.book.cash * self.size / 100 / trade_price
        else:
            raise RuntimeError("Unsupported size type")

        return asset.round_quantity(quantity), trade_price

    def pre_execute_check(
        self, ts: pd.Timestamp, trade_price: Decimal
    ) -> Optional[OrderStatus]:
        """Called with the current timestep and calculated trade price before the trade
        is executed.

        If it returns `None`, the trade is executed as normal. It can
        return `OrderStatus.CANCELLED` to indicate the trade should be
        cancelled or `OrderStatus.OPEN` to indicate the trade should not
        be executed in the current timestep and processed in the
        following timestep.
        """
        return None

    def apply(self, ts: pd.Timestamp, day_data: pd.Series, asset_map: Dict[str, Asset]):
        if not self.book or not isinstance(self.book, Book):
            raise RuntimeError("Cannot apply order without book instance")

        trade_quantity, trade_price = self._calc_quantity_price(day_data, asset_map)

        if (new_status := self.pre_execute_check(ts, trade_price)) is not None:
            self.status = new_status
            return

        trades = [
            Trade(
                asset_name=self.asset_name,
                ts=ts,
                quantity=trade_quantity,
                price=trade_price,
                order_label=self.label,
            )
        ]

        self._book_trades(trades)


class PositionalOrderCheckType(Enum):
    POS_TQ_DIFFER = 1
    ZERO_POS = 2


@mypyc_attr(allow_interpreted_subclasses=True)
@dataclass(kw_only=True)
class PositionalOrder(SimpleOrder):
    """Ensures current position is `size` and will close out existing positions to
    achieve this."""

    check_type: PositionalOrderCheckType = PositionalOrderCheckType.POS_TQ_DIFFER
    """Condition type to determine if a trade is required."""

    def __post_init__(self):
        super().__post_init__()
        self.check_type = ensure_enum(self.check_type, PositionalOrderCheckType)

    def apply(self, ts: pd.Timestamp, day_data: pd.Series, asset_map: Dict[str, Asset]):
        if not self.book or not isinstance(self.book, Book):
            raise RuntimeError("Cannot apply order without book instance")

        trade_quantity, trade_price = self._calc_quantity_price(day_data, asset_map)

        if (new_status := self.pre_execute_check(ts, trade_price)) is not None:
            self.status = new_status
            return

        current_position = self.book.positions[self.asset_name]

        if self.check_type == PositionalOrderCheckType.POS_TQ_DIFFER:
            needs_trades = current_position != trade_quantity
        elif self.check_type == PositionalOrderCheckType.ZERO_POS:
            needs_trades = current_position == 0
        else:
            raise RuntimeError(f"Unexpected check type {self.check_type}")

        trades = []

        if needs_trades:  # otherwise we're done
            if current_position:
                # close out existing position
                trades.append(
                    Trade(
                        asset_name=self.asset_name,
                        ts=ts,
                        quantity=-current_position,
                        price=trade_price,
                        order_label=self.label,
                    )
                )
            if trade_quantity != 0:
                trades.append(
                    Trade(
                        asset_name=self.asset_name,
                        ts=ts,
                        quantity=trade_quantity,
                        price=trade_price,
                        order_label=self.label,
                    )
                )

        self._book_trades(trades)


@mypyc_attr(allow_interpreted_subclasses=True)
@dataclass
class BasketOrder(Order):
    """Combine multiple assets into a single order."""

    asset_names: List[AssetName]
    """List of asset names in basket."""

    weights: List[Decimal]
    """Corresponding weights for each asset."""

    size: Decimal
    """Combined size of order."""

    size_type: OrderSizeType = OrderSizeType.QUANTITY
    """Size type."""

    def __post_init__(self):
        super().__post_init__()
        self.weights = [ensure_decimal(w) for w in self.weights]
        self.size = ensure_decimal(self.size)
        self.size_type = ensure_enum(self.size_type, OrderSizeType)

    def _calc_quantity_price(
        self, day_data: pd.Series, asset_map: Dict[str, Asset]
    ) -> List[Tuple[Decimal, Decimal]]:
        assets = [asset_map[an] for an in self.asset_names]
        assets_day_data = [a._filter_data(day_data) for a in assets]
        trade_prices = [
            asset.intraday_traded_price(add, size=self.size)
            for asset, add in zip(assets, assets_day_data)
        ]

        if self.size_type == OrderSizeType.QUANTITY:
            quantities = [self.size * w for w in self.weights]
        elif self.size_type == OrderSizeType.NOTIONAL:
            # size = k * sum(w_i * p_i)
            tp_weighted = sum(w * p for w, p in zip(self.weights, trade_prices))
            k = self.size / tp_weighted
            quantities = [k * w for w in self.weights]
        elif self.size_type == OrderSizeType.BOOK_PERCENT:
            assert isinstance(self.book, Book)  # to please mypy
            # TODO: size is ignored, perhaps use a scaling factor?
            # NOTE: we could use self.book.mtm but would be from previous day
            book_mtm = sum(
                [
                    self.book.positions.get(a.name, 0) * tp
                    for a, tp in zip(assets, trade_prices)
                ]
            )
            book_value = self.book.cash + book_mtm
            quantities = [
                book_value * w / 100 / tp for w, tp in zip(self.weights, trade_prices)
            ]
        else:
            raise RuntimeError("Unsupported size type")

        return [
            (a.round_quantity(q), tp)
            for a, q, tp in zip(assets, quantities, trade_prices)
        ]

    def apply(self, ts: pd.Timestamp, day_data: pd.Series, asset_map: Dict[str, Asset]):
        if not self.book or not isinstance(self.book, Book):
            raise RuntimeError("Cannot apply order without book instance")

        trade_quantity_prices = self._calc_quantity_price(day_data, asset_map)

        trades = [
            Trade(
                asset_name=an,
                ts=ts,
                quantity=tq,
                price=tp,
                order_label=self.label,
            )
            for an, (tq, tp) in zip(self.asset_names, trade_quantity_prices)
        ]

        self._book_trades(trades)


@mypyc_attr(allow_interpreted_subclasses=True)
@dataclass(kw_only=True)
class PositionalBasketOrder(BasketOrder):
    """Similar to a :py:class:`BasketOrder` but will close out existing positions if
    they do not match requested weights."""

    check_type: PositionalOrderCheckType = PositionalOrderCheckType.POS_TQ_DIFFER

    def apply(self, ts: pd.Timestamp, day_data: pd.Series, asset_map: Dict[str, Asset]):
        if not self.book or not isinstance(self.book, Book):
            raise RuntimeError("Cannot apply order without book instance")

        trade_quantity_prices = self._calc_quantity_price(day_data, asset_map)

        current_positions = [self.book.positions[an] for an in self.asset_names]

        if self.check_type == PositionalOrderCheckType.POS_TQ_DIFFER:
            needs_trades = any(
                p != tq for p, (tq, tp) in zip(current_positions, trade_quantity_prices)
            )
        elif self.check_type == PositionalOrderCheckType.ZERO_POS:
            needs_trades = any(p == 0 for p in current_positions)
        else:
            raise RuntimeError(f"Unexpected check type {self.check_type}")

        trades = []

        if needs_trades:  # otherwise we're done
            for asset_name, current_position, (trade_quantity, trade_price) in zip(
                self.asset_names, current_positions, trade_quantity_prices
            ):
                if current_position:
                    # close out existing position
                    trades.append(
                        Trade(
                            asset_name=asset_name,
                            ts=ts,
                            quantity=-current_position,
                            price=trade_price,
                            order_label=self.label,
                        )
                    )
                if trade_quantity != 0:
                    trades.append(
                        Trade(
                            asset_name=asset_name,
                            ts=ts,
                            quantity=trade_quantity,
                            price=trade_price,
                            order_label=self.label,
                        )
                    )

        self._book_trades(trades)



import unittest

from yabte.tests._helpers import notebooks_dir

HAS_NBFORMAT = True
try:
    import nbformat
    from nbconvert.preprocessors import ExecutePreprocessor
except:
    HAS_NBFORMAT = False


class NotebooksTestCase(unittest.TestCase):
    @unittest.skipUnless(HAS_NBFORMAT, "needs nbformat")
    def test_notebooks_smoke(self):
        for nb in notebooks_dir.glob("*.ipynb"):
            if not nb.name.startswith("_"):
                with self.subTest(nb.name):
                    ep = ExecutePreprocessor(timeout=600)
                    with nb.open() as f:
                        nb = nbformat.read(f, as_version=4)
                        # will raise exception if issue, eg missing module
                        nb_node, resources = ep.preprocess(
                            nb, {"metadata": {"path": str(notebooks_dir)}}
                        )
                        foo = 1



r"""Calculate portfolio weights by inverting variance.

That is,

.. math::

   w = \frac{\sigma^{-1}}{1'\sigma^{-1}}

where :math:`\sigma^2 = diag(\Sigma)`.
"""

import numpy as np


def inverse_volatility(cov: np.ndarray) -> np.ndarray:
    """Calculate weights using inverse variance."""
    sigma_inv = np.sqrt(np.diag(cov)) ** -1
    return sigma_inv / (np.ones_like(sigma_inv) @ sigma_inv)



from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal
from typing import TYPE_CHECKING, TypeAlias, TypeVar, Union, cast

import pandas as pd
from mypy_extensions import mypyc_attr

__all__ = ["OHLCAsset"]


# use ints until mypyc supports IntFlag
# https://github.com/mypyc/mypyc/issues/1022
AssetDataFieldInfo = int
ADFI_AVAILABLE_AT_CLOSE: int = 1
ADFI_AVAILABLE_AT_OPEN: int = 2
ADFI_REQUIRED: int = 4

T = TypeVar("T", bound=Union[pd.Series, pd.DataFrame])

AssetName: TypeAlias = str
"""Asset name string."""


@mypyc_attr(allow_interpreted_subclasses=True)
@dataclass(kw_only=True)
class Asset:
    """Anything that has a price."""

    name: AssetName
    """Name string."""

    denom: str = "USD"
    """Denominated currency."""

    price_round_dp: int = 2
    """Number of decimal places to round prices to."""

    quantity_round_dp: int = 2
    """Number of decimal places to round quantities to."""

    data_label: str | None = None
    """`StrategyRunner.data` column index 1st level label.

    Defaults to `name`
    """

    def __post_init__(self):
        if self.data_label is None:
            self.data_label = self.name

    def round_quantity(self, quantity) -> Decimal:
        """Round `quantity`."""
        return round(quantity, self.quantity_round_dp)

    def intraday_traded_price(
        self, asset_day_data: pd.Series, size: Decimal | None = None
    ) -> Decimal:
        """Calculate price during market hours with given row of `asset_day_data` and
        the order `size`.

        The `size` can be used to
        determine a price from say, bid / ask spreads.
        """
        raise NotImplementedError(
            "The intraday_traded_price method needs to be implemented."
        )

    def end_of_day_price(self, asset_day_data: pd.Series) -> Decimal:
        """Calculate price at end of day with given row of `asset_day_data`."""
        raise NotImplementedError(
            "The end_of_day_price method needs to be implemented."
        )

    def check_and_fix_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Checks dataframe `data` has correct fields and fixes columns where
        necessary."""
        raise NotImplementedError(
            "The check_and_fix_data method needs to be implemented."
        )

    def data_fields(self) -> list[tuple[str, AssetDataFieldInfo]]:
        """List of data fields and their availability."""
        raise NotImplementedError("The data_fields method needs to be implemented.")

    def _get_fields(self, field_info: AssetDataFieldInfo) -> list[str]:
        """Internal method to get fields from `data_fields` with `field_info`."""
        return [f for f, fi in self.data_fields() if fi & field_info]

    def _filter_data(self, data: T) -> T:
        """Internal method to filter `data` columns and return only those relevant to
        pricing."""
        assert isinstance(self.data_label, str)
        return cast(T, data[self.data_label])


@mypyc_attr(allow_interpreted_subclasses=True)
@dataclass(kw_only=True)
class OHLCAsset(Asset):
    """Assets whose price history is represented by High, Low, Open, Close and Volume
    fields."""

    def data_fields(self) -> list[tuple[str, AssetDataFieldInfo]]:
        return [
            ("High", ADFI_AVAILABLE_AT_CLOSE),
            ("Low", ADFI_AVAILABLE_AT_CLOSE),
            ("Open", ADFI_AVAILABLE_AT_CLOSE | ADFI_AVAILABLE_AT_OPEN),
            ("Close", ADFI_AVAILABLE_AT_CLOSE | ADFI_REQUIRED),
            ("Volume", ADFI_AVAILABLE_AT_CLOSE),
        ]

    def intraday_traded_price(
        self, asset_day_data: pd.Series, size: Decimal | None = None
    ) -> Decimal:
        if pd.notnull(asset_day_data.Low) and pd.notnull(asset_day_data.High):
            p = Decimal((asset_day_data.Low + asset_day_data.High) / 2)
        else:
            p = Decimal(asset_day_data.Close)
        return round(p, self.price_round_dp)

    def end_of_day_price(self, asset_day_data: pd.Series) -> Decimal:
        return round(Decimal(asset_day_data.Close), self.price_round_dp)

    def check_and_fix_data(self, data: pd.DataFrame) -> pd.DataFrame:
        # TODO: check low <= open, high, close & high >= open, low, close
        # TODO: check volume >= 0

        # check each asset has required fields
        required_fields = self._get_fields(ADFI_REQUIRED)
        missing_req_fields = set(required_fields) - set(data.columns)
        if len(missing_req_fields):
            raise ValueError(
                f"data columns index requires fields {required_fields}"
                f" and missing {missing_req_fields}"
            )

        # reindex columns with expected fields + additional fields
        expected_fields = self._get_fields(ADFI_AVAILABLE_AT_CLOSE)
        other_fields = list(set(data.columns) - set(expected_fields))
        return data.reindex(expected_fields + other_fields, axis=1)



r"""Calculate portfolio weights with hierarchical risk parity.

That is to employ hierarchical tree clustering on the correlation distance
matrix and quasi-diagonalisation followed by recursive bisection to determine
the weights. See [LP]_ for further details.


References
----------
.. [LP] López de Prado, M. (2016). Building Diversified Portfolios that
    Outperform Out of Sample. The Journal of Portfolio Management, 42(4),
    59–69. https://doi.org/10.3905/jpm.2016.42.4.059
"""

import numpy as np
import pandas as pd
from scipy.cluster.hierarchy import linkage, to_tree

# following 3 functions taken directly from paper [LP]


def _getIVP(cov, **kargs):
    # Compute the inverse-variance portfolio
    ivp = 1.0 / np.diag(cov)
    ivp /= ivp.sum()
    return ivp


def _getClusterVar(cov, cItems):
    # Compute variance per cluster
    cov_ = cov.loc[cItems, cItems]  # matrix slice
    w_ = _getIVP(cov_).reshape(-1, 1)
    cVar = np.dot(np.dot(w_.T, cov_), w_)[0, 0]
    return cVar


def _getRecBipart(cov, sortIx):
    # Compute HRP alloc
    w = pd.Series(1.0, index=sortIx)
    cItems = [sortIx]  # initialize all items in one cluster
    while len(cItems) > 0:
        cItems = [
            i[j:k]
            for i in cItems
            for j, k in ((0, len(i) // 2), (len(i) // 2, len(i)))
            if len(i) > 1
        ]  # bi-section
        for i in range(0, len(cItems), 2):  # parse in pairs
            cItems0 = cItems[i]  # cluster 1
            cItems1 = cItems[i + 1]  # cluster 2
            cVar0 = _getClusterVar(cov, cItems0)
            cVar1 = _getClusterVar(cov, cItems1)
            alpha = 1 - cVar0 / (cVar0 + cVar1)
            w[cItems0] *= alpha  # weight 1
            w[cItems1] *= 1 - alpha  # weight 2
    return w


def hrp(corr: pd.DataFrame, sigma: np.ndarray) -> np.ndarray:
    """Calculate weights using hierarchical risk parity and scipy's linkage/to_tree
    functions."""
    cov = np.diag(sigma) @ corr @ np.diag(sigma)
    cov.index, cov.columns = corr.index, corr.columns
    rho = corr.values
    D = np.sqrt((1 - rho) / 2)
    I, J = np.triu_indices_from(D, 1)
    link = linkage(np.sqrt(np.sum((D[I] - D[J]) ** 2, axis=1)))
    ix_sorted = to_tree(link, rd=False).pre_order()
    cols_sorted = corr.columns[ix_sorted]
    return _getRecBipart(cov, cols_sorted)



import logging
from collections import defaultdict
from dataclasses import dataclass, field
from decimal import Decimal
from itertools import groupby
from typing import Any, Dict, List, Sequence, TypeAlias

import pandas as pd

from ._helpers import ensure_decimal
from .asset import Asset, AssetName
from .transaction import CashTransaction, Trade, Transaction

logger = logging.getLogger(__name__)


__all__ = ["Book"]


@dataclass(kw_only=True)
class BookMandate:
    def check(self, current_pos, quantity) -> bool:
        raise NotImplementedError()


BookName: TypeAlias = str
"""Book name string."""


@dataclass(kw_only=True)
class Book:
    """Record of asset trades and positions including cash.

    A `name` can be provided and the default `cash` value can be changed
    to a non-zero amount.
    """

    name: BookName
    """Name of book."""

    mandates: Dict[AssetName, BookMandate] = field(default_factory=dict)
    """Dictionary mapping assets to mandates (experimental)."""

    positions: Dict[AssetName, Decimal] = field(
        default_factory=lambda: defaultdict(Decimal)
    )
    """Dictionary tracking positions of each asset."""

    transactions: List[Transaction] = field(default_factory=list)
    """List of executed trades."""

    denom: str = "USD"
    """Book currency."""

    cash: Decimal = Decimal(0)
    """Cash value of book."""

    rate: Decimal = Decimal(0)
    """Daily interest rate applied to cash at end of day."""

    interest_round_dp: int = 3
    """Number of decimal places to round interest."""

    _history: List[List[Any]] = field(default_factory=list)

    @property
    def history(self) -> pd.DataFrame:
        """Dataframe with book cash, mtm and total value history."""
        return pd.DataFrame(
            self._history, columns=["ts", "cash", "mtm", "total"]
        ).set_index("ts")

    def __post_init__(self):
        self.cash = ensure_decimal(self.cash)
        self.rate = ensure_decimal(self.rate)

    def test_trades(self, trades: Sequence[Trade]) -> bool:
        """Checks whether list of trades will be successful by not failing any
        mandates."""
        for asset_name, asset_trades in groupby(trades, lambda t: t.asset_name):
            if asset_name in self.mandates:
                total_quantity = sum(t.quantity for t in asset_trades)
                if not self.mandates[asset_name].check(
                    self.positions[asset_name], total_quantity
                ):
                    return False
        return True

    def add_transactions(self, transactions: Sequence[Transaction]):
        """Records the `transactions` and adjusts internal dictionary of positions and
        value of cash accordingly."""
        for tran in transactions:
            if isinstance(tran, Trade):
                self.positions[tran.asset_name] += tran.quantity
                self.cash += tran.total
            elif isinstance(tran, CashTransaction):
                self.cash += tran.total
            else:
                raise ValueError(f"Unsupport transaction class: {type(tran)}")

            self.transactions.append(tran)

    def eod_tasks(
        self, ts: pd.Timestamp, day_data: pd.Series, asset_map: Dict[str, Asset]
    ):
        """Run end of day tasks such as book keeping."""
        # accumulate continously compounded interest
        interest = round(self.cash * (self.rate.exp() - 1), self.interest_round_dp)
        if self.rate != 0 and interest != 0:
            self.add_transactions(
                [
                    CashTransaction(
                        ts=ts,
                        total=interest,
                        desc=f"interest payment on cash {self.cash:.2f}",
                    )
                ]
            )
        cash = float(self.cash)
        mtm = float(
            sum(
                asset.end_of_day_price(asset._filter_data(day_data)) * q
                for an, q in self.positions.items()
                if (asset := asset_map.get(an))
            )
        )
        self._history.append([ts, cash, mtm, cash + mtm])



import logging
from decimal import Decimal
from enum import Enum
from typing import Any, Type, TypeVar

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Enum)


def ensure_enum(value: Any, enum_type: Type[T]) -> T:
    if isinstance(value, enum_type):
        return value
    if isinstance(value, str):
        return enum_type[value.upper()]
    if isinstance(value, int):
        return enum_type(value)
    raise ValueError(f"Unexpected enum type {value} for {enum_type}")


def ensure_decimal(value: Any) -> Decimal:
    if isinstance(value, Decimal):
        return value
    if isinstance(value, (str, float, int)):
        return Decimal(value)
    raise ValueError(f"Unexpected decimal type {value}")



import logging
from dataclasses import dataclass
from itertools import chain, product
from typing import Dict

import pandas as pd
from mypy_extensions import mypyc_attr

# TODO: use explicit imports until mypyc fixes attribute lookups in dataclass
# (https://github.com/mypyc/mypyc/issues/1000)
from pandas import DataFrame, Series, Timestamp  # type: ignore

from .asset import ADFI_AVAILABLE_AT_OPEN, Asset, AssetName
from .book import Book, BookName
from .order import Orders

logger = logging.getLogger(__name__)

__all__ = ["Strategy"]


@mypyc_attr(allow_interpreted_subclasses=True)
@dataclass(kw_only=True)
class Strategy:
    """Trading strategy base class.

    This class should be derived from and will be instantiated by
    :py:class:`StrategyRunner`.
    """

    orders: Orders | None = None
    """Double ended queue of orders."""

    params: pd.Series | None = None
    """Parameters supplied to strategy."""

    books: Dict[BookName, Book] | None = None
    """Dictionary of books."""

    assets: Dict[AssetName, Asset] | None = None
    """Dictionary of assets."""

    _ts: pd.Timestamp | None = None
    _data_lock: bool = True
    _mask_open: bool = False

    @property
    def ts(self) -> pd.Timestamp | None:
        """Stores the current timestamp."""
        return self._ts

    def _set_ts(self, ts: pd.Timestamp):
        """Internal method to update timestep to current `ts`"""
        self._ts = ts

    def _get_col_indexer(self):
        # cache this call for hopefully a small speed up
        if not hasattr(self, "_col_indexer"):
            mix = pd.MultiIndex.from_tuples(
                chain(
                    *[
                        product(
                            [asset.data_label],
                            asset._get_fields(ADFI_AVAILABLE_AT_OPEN),
                        )
                        for asset_name, asset in self.assets.items()
                    ]
                )
            )
            self._col_indexer = ~self._data.columns.isin(mix)
        return self._col_indexer

    @property
    def data(self) -> pd.DataFrame:
        """Provides window of data available up to current timestamp `self.ts` and masks
        out data not availble at open (like high, low, close, volume)."""
        if not self.ts:
            return self._data
        else:
            df_t = self._data.loc[: self.ts, :]  # type: ignore[misc]
            if not self._mask_open:
                data = df_t
            else:
                row_indexer = df_t.index == df_t.index[-1]
                # generate mask from asset instances, some assets
                # might support different field masks
                col_indexer = self._get_col_indexer()
                mask = row_indexer[:, None] & col_indexer
                data = df_t.mask(mask)
            return data

    @data.setter
    def data(self, value: pd.DataFrame):
        if self._data_lock:
            raise RuntimeError("Attempt to write data post init")
        self._data = value

    def init(self):
        """Initialise internal variables & enhance data for strategy."""
        pass

    def on_open(self):
        """Executed on open every day.

        Use `self.ts` to determine current timestamp. Data available at
        this timestamp is accessible from `self.data`.
        """
        pass

    def on_close(self):
        """Executed on close every day.

        Use `self.ts` to determine current timestamp. Data available at
        this timestamp is accessible from `self.data`.
        """
        pass



r"""Geometric Brownian motion simulation.

Simulate stochastic process :math:`S_t` where,

.. math::

   dS_t = \mu S_t dt + \sigma S_t dW_t

where :math:`\mu` is the drift, :math:`\sigma` is the volatility
and :math:`dW_t` is a Weiner process.
"""

import numpy as np

from yabte.utilities.simulation.weiner import weiner_simulate_paths


def gbm_simulate_paths(
    S0: float,
    mu: np.ndarray,
    sigma: np.ndarray,
    R: np.ndarray,
    T: int,
    n_steps: int,
    n_sims: int,
    rng=None,
):
    """Generate simulated paths using vectorised numpy calls.

    `S0` is initial value, `mu` is the drift, `sigma` is volatility, `R` a correlation
    matrix, `T` is the time span, `n_steps` is how many time steps,
    `n_sims` the number of simulations and `rng` a numpy random
    number generator (optional).

    TODO: support Euler–Maruyama / Milstein / Antithetic.
    """

    mu = np.atleast_1d(mu)
    sigma = np.atleast_1d(sigma)

    if rng is None:
        rng = np.random.default_rng()

    r = mu  # mu = rf in risk neutral framework
    dt = T / n_steps

    # duplicate copies of time axis to simplify broadcasting later
    ts = np.linspace(0, T, n_steps, endpoint=False)
    ts = np.repeat(ts[:, np.newaxis], n_sims, axis=1)[:, :, np.newaxis]

    ws = weiner_simulate_paths(
        n_steps=n_steps, n_sims=n_sims, stdev=np.sqrt(dt), R=R, rng=rng
    )

    # use closed form solution
    return S0 * np.exp((r - sigma**2 / 2) * ts + sigma * ws)



import unittest

import numpy.testing as nptu

__all__ = ["NumpyTestCase"]


class NumpyTestCase(unittest.TestCase):
    """Specialized TestCase which includes numpy test assertion functions and maps them
    from assert_func_name to self.numpyAssertFuncName.

    class NumpyTest(NumpyTestCase):
        def test_allclose_example(self):
            a1 = np.array([1.,2.,3.])
            self.numpyAssertAllclose(a1, np.array([1.,2.,3.1]))
    """


def make_test_wrapper(fn):
    def test_wrapper(self, *args, **kwargs):
        try:
            nptu.__dict__[fn](*args, **kwargs)
        except AssertionError as err:
            self.fail(err)

    return test_wrapper


for fn in nptu.__dict__:
    if fn.startswith("assert") and not fn.endswith("_"):
        new_name = "numpy" + fn.title().replace("_", "")
        setattr(NumpyTestCase, new_name, make_test_wrapper(fn))


if __name__ == "__main__":
    import numpy as np

    class _NumpyTest(NumpyTestCase):
        def test_allclose_example(self):
            self.numpyAssertAllclose(
                np.array([1.0, 2.0, 3.0]), np.array([1.0, 2.0, 3.1])
            )

    # should fail!
    suite = unittest.TestSuite()
    suite.addTest(_NumpyTest("test_allclose_example"))
    unittest.TextTestRunner().run(suite)



from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.dates import DateFormatter
from matplotlib.ticker import FuncFormatter

from ....backtest import StrategyRunner, StrategyRunnerResult
from .marker_updater import MarkerUpdater


def plot_strategy_runner_result(
    srr: StrategyRunnerResult,
    sr: StrategyRunner,
    settings: dict[str, Any] | None = None,
):
    """Display the results of a strategy run using matplotlib.

    Plots a grid of charts with each column representing a book and rows representing
    each asset's price series along with long/short positioning and volumne series. A
    bottom row shows the value of each book as a price series.
    """
    default_settings = {
        "candle_body_width": 0.8,
        "candle_wick_width": 0.2,
        "candle_up_color": "green",
        "candle_down_color": "red",
        "date_format": "%d-%b-%Y",
        "long_markers": {},
        "short_markers": {},
    }

    if isinstance(settings, dict):
        default_settings = default_settings | settings
    s = pd.Series(default_settings, dtype=object)

    traded_assets = [
        a for a in srr.assets if a.name in srr.transaction_history.asset_name.unique()
    ]

    col_width = 8
    row_unit_height = 3
    ncols = len(srr.books)
    nrows = 1 + 2 * len(traded_assets)
    hr = [3, 1] * len(traded_assets) + [2]
    fig, axss = plt.subplots(
        nrows,
        ncols,
        height_ratios=hr,
        sharex=True,
        squeeze=False,
        figsize=(ncols * col_width, nrows * row_unit_height),
    )
    fig.suptitle("Strategy Runner Report")

    marker_updater = MarkerUpdater()
    date_formatter = DateFormatter(s.date_format)

    for book, axs in zip(srr.books, axss.T):
        for i, asset in enumerate(traded_assets):
            prices = asset._filter_data(sr.data)

            up = prices[prices.Close >= prices.Open]
            down = prices[prices.Close < prices.Open]

            pax = axs[2 * i]
            pax.set_title(asset.name)
            pax.set_ylabel(asset.denom)
            pax.fmt_xdata = date_formatter

            pax.bar(
                up.index,
                up.Close - up.Open,
                s.candle_body_width,
                bottom=up.Open,
                color=s.candle_up_color,
            )
            pax.bar(
                up.index,
                up.High - up.Close,
                s.candle_wick_width,
                bottom=up.Close,
                color=s.candle_up_color,
            )
            pax.bar(
                up.index,
                up.Low - up.Open,
                s.candle_wick_width,
                bottom=up.Open,
                color=s.candle_up_color,
            )

            pax.bar(
                down.index,
                down.Close - down.Open,
                s.candle_body_width,
                bottom=down.Open,
                color=s.candle_down_color,
            )
            pax.bar(
                down.index,
                down.High - down.Open,
                s.candle_wick_width,
                bottom=down.Open,
                color=s.candle_down_color,
            )
            pax.bar(
                down.index,
                down.Low - down.Close,
                s.candle_wick_width,
                bottom=down.Close,
                color=s.candle_down_color,
            )

            vax = axs[2 * i + 1]
            vax.yaxis.set_major_formatter(FuncFormatter(lambda x, p: f"{x/1000:,.0f}"))
            vax.set_ylabel("Volume (thousands)")
            vax.fmt_xdata = date_formatter

            vax.bar(
                prices.index,
                prices.Volume,
                color="black",
            )

            trans_hist = srr.transaction_history.query(
                "asset_name==@asset.name and book==@book.name"
            )
            pos_hist = (
                trans_hist.groupby("ts")
                .agg(
                    quantity=("quantity", np.sum),
                    labels=(
                        "order_label",
                        lambda L: " ".join(l for l in L if l is not None),
                    ),
                )
                .reindex(prices.index)
            )

            for l in pos_hist.labels.dropna().unique():
                marker_long = s.long_markers.get("", "^")
                marker_short = s.long_markers.get("", "v")

                prices.Low[pos_hist.eval("quantity < 0 and labels==@l")].rename(
                    f"Short {l}".strip()
                ).plot(
                    color="red",
                    marker=marker_short,
                    markersize=6,
                    linestyle="None",
                    ax=pax,
                )
                prices.High[pos_hist.eval("quantity > 0 and labels==@l")].rename(
                    f"Long {l}".strip()
                ).plot(
                    color="green",
                    marker=marker_long,
                    markersize=6,
                    linestyle="None",
                    ax=pax,
                )

            pax.legend()
            marker_updater.add_ax(pax, ["size"])

        bax = axs[-1]
        bax.yaxis.set_major_formatter(FuncFormatter(lambda x, p: f"{x/1000:,.0f}"))
        bax.set_ylabel(f"{book.denom} (thousands)")

        srr.book_history.loc[:, book.name].total.plot(ax=bax)
        bax.set_xlabel("Date")
        bax.fmt_xdata = date_formatter

    # markers don't scale with zoom, so update them in this event hook
    fig.tight_layout()
    return fig, axs



import numpy as np
import pandas as pd


@pd.api.extensions.register_dataframe_accessor("scl")
class ScaleAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    @property
    def standard(self):
        return (self._obj - self._obj.mean()) / self._obj.std()


@pd.api.extensions.register_dataframe_accessor("prc")
@pd.api.extensions.register_series_accessor("prc")
class PriceAccessor:
    # TODO: add ledoit cov (via sklearn)
    # http://www.ledoit.net/honey.pdf

    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj):
        pass

    @property
    def log_returns(self):
        if (self._obj < 0).any(axis=None):
            raise AttributeError("Prices must be non-negative for log returns")
        return np.log((self._obj / self._obj.shift())[1:])

    @property
    def returns(self):
        return self._obj.pct_change(fill_method=None)[1:]

    @property
    def frequency(self):
        days = pd.Timedelta(np.diff(self._obj.index).min()).days
        if days == 1:
            return 252
        elif days == 7:
            return 52
        elif 28 <= days <= 31:
            return 12

    def capm_returns(self, risk_free_rate=0):
        returns = self.returns
        returns_mkt = returns.mean(axis=1).rename("MKT")

        # concat market retutns & compute sampel covariance matrix
        cov = pd.concat([returns, returns_mkt], axis=1).cov()
        betas = cov.MKT.drop("MKT") / cov.MKT.MKT

        return (
            risk_free_rate
            + betas * (returns_mkt.mean() * self.frequency - risk_free_rate)
        ).rename("CAPM")

    def sharpe_ratio(self, risk_free_rate=0, use_log_returns=True):
        ann_factor = np.sqrt(self.frequency)
        returns = self.log_returns if use_log_returns else self.returns
        return ann_factor * (returns.mean() - risk_free_rate) / returns.std()

    def null_blips(self, sd=5, sdd=7):
        df = self._obj
        z = df.scl.standard
        zd = z.diff()
        # TODO support blips longer than 1 day?
        for col, series in df[z.abs() > sd].dropna(how="all").dropna(axis=1).items():
            for row, val in series.items():
                row_ix = df.index.get_loc(row)
                col_ix = df.columns.get_loc(col)
                zd0 = zd.iloc[row_ix, col_ix]
                zd1 = zd.iloc[row_ix + 1, col_ix]
                if (
                    np.sign(zd0) != np.sign(zd1)
                    and np.abs(zd0) > sdd
                    and np.abs(zd1) > sdd
                ):
                    print(f"nullifying blip at {col} {row}")
                    df.iloc[row_ix, col_ix] = np.nan

        return self._obj



from typing import Any

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from ....backtest import StrategyRunner, StrategyRunnerResult


def plot_strategy_runner_result(
    srr: StrategyRunnerResult,
    sr: StrategyRunner,
    settings: dict[str, Any] | None = None,
):
    """Display the results of a strategy run using plotly.

    Plots a grid of charts with each column representing a book and rows representing
    each asset's price series along with long/short positioning and volumne series. A
    bottom row shows the value of each book as a price series.
    """
    default_settings: dict[str, Any] = {}

    if isinstance(settings, dict):
        default_settings = default_settings | settings
    s = pd.Series(default_settings, dtype=object)

    traded_assets = [
        a for a in srr.assets if a.name in srr.transaction_history.asset_name.unique()
    ]

    dpi = 100
    col_width = 8 * dpi
    row_unit_height = 3 * dpi
    ncols = len(srr.books)
    nrows = 1 + len(traded_assets)

    subplot_titles = [a.name for a in (traded_assets)] + ["Book Value"]
    row_heights = [10 for a in (traded_assets)] + [2]
    specs = [[{"secondary_y": True} for c in range(ncols)] for r in range(nrows)]

    fig = make_subplots(
        rows=nrows,
        cols=ncols,
        shared_xaxes=True,
        subplot_titles=subplot_titles,
        row_heights=row_heights,
        start_cell="top-left",
        specs=specs,
        vertical_spacing=0.05,
    )

    for col, book in enumerate(srr.books, start=1):
        for row, asset in enumerate(traded_assets, start=1):
            prices = asset._filter_data(sr.data)

            fig.add_trace(
                go.Candlestick(
                    x=prices.index,
                    open=prices.Open,
                    high=prices.High,
                    low=prices.Low,
                    close=prices.Close,
                ),
                row=row,
                col=col,
                secondary_y=True,
            )

            fig.add_trace(
                go.Bar(
                    x=prices.index,
                    y=prices.Volume,
                    marker_color="lightgrey",
                ),
                row=row,
                col=col,
                secondary_y=False,
            )

            fig.update_xaxes(rangeslider_visible=False, row=row, col=col)
            fig.update_yaxes(
                title=asset.denom, secondary_y=True, showgrid=True, row=row, col=col
            )
            fig.update_yaxes(
                title="Volume", secondary_y=False, showgrid=False, row=row, col=col
            )

            # add some range selector buttons
            fig.update_xaxes(
                rangeselector=dict(
                    buttons=[
                        dict(count=1, label="1m", step="month", stepmode="backward"),
                        dict(count=6, label="6m", step="month", stepmode="backward"),
                        dict(count=1, label="YTD", step="year", stepmode="todate"),
                        dict(count=1, label="1y", step="year", stepmode="backward"),
                        dict(step="all"),
                    ]
                ),
                row=1,
                col=col,
            )

            trans_hist = srr.transaction_history.query(
                "asset_name==@asset.name and book==@book.name"
            )
            pos_hist = (
                trans_hist.groupby("ts")
                .agg(
                    quantity=("quantity", "sum"),
                    labels=(
                        "order_label",
                        lambda L: " ".join(l for l in L if l is not None),
                    ),
                )
                .reindex(prices.index)
            )

            shorts = prices[pos_hist.eval("quantity < 0")][["Low"]].join(
                pos_hist.labels
            )
            fig.add_trace(
                go.Scatter(
                    x=shorts.index,
                    y=shorts.Low,
                    customdata=shorts.labels,
                    mode="markers",
                    marker_symbol="arrow-down",
                    marker_color="red",
                    marker_size=10,
                    hovertemplate="%{x}<br>%{y}<br>%{customdata}<extra></extra>",
                ),
                row=row,
                col=col,
                secondary_y=True,
            )

            longs = prices[pos_hist.eval("quantity > 0")][["High"]].join(
                pos_hist.labels
            )
            fig.add_trace(
                go.Scatter(
                    x=longs.index,
                    y=longs.High,
                    customdata=shorts.labels,
                    mode="markers",
                    marker_symbol="arrow-up",
                    marker_color="green",
                    marker_size=10,
                    hovertemplate="%{x}<br>%{y}<br>%{customdata}<extra></extra>",
                ),
                row=row,
                col=col,
                secondary_y=True,
            )

    row = nrows
    bh = srr.book_history.loc[:, book.name]
    fig.add_trace(
        go.Scatter(
            x=bh.index,
            y=bh.total,
        ),
        row=row,
        col=col,
    )

    fig.update_xaxes(rangeslider_visible=True, row=row, col=col)
    fig.update_yaxes(title=book.denom, showgrid=True, row=row, col=col)
    fig.update_xaxes(title="Date", row=row, col=col)

    fig.update_layout(
        height=nrows * row_unit_height,
        width=ncols * col_width,
        showlegend=False,
        title_text="Strategy Runner Report",
    )

    return fig



class MarkerUpdater:
    """Allow updating marker sizes & alphas on zoom.

    The original is from SO by Thomas Kühn.

    https://stackoverflow.com/a/48475557/370696
    """

    def __init__(self):
        ##for storing information about Figures and Axes
        self.figs = {}

        ##for storing timers
        self.timer_dict = {}

    def add_ax(self, ax, features=[]):
        ax_dict = self.figs.setdefault(ax.figure, dict())
        ax_dict[ax] = {
            "xlim": ax.get_xlim(),
            "ylim": ax.get_ylim(),
            "figw": ax.figure.get_figwidth(),
            "figh": ax.figure.get_figheight(),
            "scale_s": 1.0,
            "scale_a": 1.0,
            "features": [features] if isinstance(features, str) else features,
        }
        ax.figure.canvas.mpl_connect("draw_event", self.update_axes)

    def update_axes(self, event):
        for fig, axes in self.figs.items():
            if fig is event.canvas.figure:
                for ax, args in axes.items():
                    ##make sure the figure is re-drawn
                    update = True

                    fw = fig.get_figwidth()
                    fh = fig.get_figheight()
                    fac1 = min(fw / args["figw"], fh / args["figh"])

                    xl = ax.get_xlim()
                    yl = ax.get_ylim()
                    fac2 = min(
                        abs(args["xlim"][1] - args["xlim"][0]) / abs(xl[1] - xl[0]),
                        abs(args["ylim"][1] - args["ylim"][0]) / abs(yl[1] - yl[0]),
                    )

                    ##factor for marker size
                    facS = (fac1 * fac2) / args["scale_s"]

                    ##factor for alpha -- limited to values smaller 1.0
                    facA = min(1.0, fac1 * fac2) / args["scale_a"]

                    ##updating the artists
                    if facS != 1.0:
                        for line in ax.lines:
                            if "size" in args["features"]:
                                line.set_markersize(line.get_markersize() * facS)

                            if "alpha" in args["features"]:
                                alpha = line.get_alpha()
                                if alpha is not None:
                                    line.set_alpha(alpha * facA)

                        for path in ax.collections:
                            if "size" in args["features"]:
                                path.set_sizes(
                                    [s * facS**2 for s in path.get_sizes()]
                                )

                            if "alpha" in args["features"]:
                                alpha = path.get_alpha()
                                if alpha is not None:
                                    path.set_alpha(alpha * facA)

                        args["scale_s"] *= facS
                        args["scale_a"] *= facA

                self._redraw_later(fig)

    def _redraw_later(self, fig):
        timer = fig.canvas.new_timer(interval=10)
        timer.single_shot = True
        timer.add_callback(lambda: fig.canvas.draw_idle())
        timer.start()

        ##stopping previous timer
        if fig in self.timer_dict:
            self.timer_dict[fig].stop()

        ##storing a reference to prevent garbage collection
        self.timer_dict[fig] = timer



# TODO: use absolute imports until mypyc fixes relative imports in __init__.py
# (https://github.com/mypyc/mypyc/issues/996)
from yabte.backtest.asset import (
    ADFI_AVAILABLE_AT_CLOSE,
    ADFI_AVAILABLE_AT_OPEN,
    ADFI_REQUIRED,
    AssetDataFieldInfo,
    AssetName,
    OHLCAsset,
)
from yabte.backtest.book import Book, BookMandate, BookName
from yabte.backtest.order import (
    BasketOrder,
    OrderSizeType,
    OrderStatus,
    PositionalBasketOrder,
    PositionalOrder,
    PositionalOrderCheckType,
    SimpleOrder,
)
from yabte.backtest.strategy import Strategy
from yabte.backtest.strategyrunner import StrategyRunner, StrategyRunnerResult
from yabte.backtest.transaction import CashTransaction, Trade

__all__ = [
    "OHLCAsset",
    "AssetDataFieldInfo",
    "AssetName",
    "Book",
    "BookName",
    "BookMandate",
    "CashTransaction",
    "SimpleOrder",
    "OrderSizeType",
    "OrderStatus",
    "BasketOrder",
    "PositionalBasketOrder",
    "PositionalOrder",
    "PositionalOrderCheckType",
    "Trade",
    "Strategy",
    "StrategyRunner",
    "StrategyRunnerResult",
    "ADFI_AVAILABLE_AT_CLOSE",
    "ADFI_AVAILABLE_AT_OPEN",
    "ADFI_REQUIRED",
]



r"""Calculate portfolio weights by minimizing variance.

That is to minimize the following expression subject to various constraints.

.. math::

   \frac{1}{2} w' \Sigma w

Typical constraints are achieving a target return, i.e. :math:`w' \mu = r`, and that
all weights sum to one, i.e. :math:`\Sigma_i w_i = 1`.
"""

import numpy as np
import numpy.linalg as la

from ..lagrangian import Lagrangian


def minimum_variance(Sigma: np.ndarray, mu: np.ndarray, r: float) -> np.ndarray:
    """Calculate weights using Lagrangian multipliers and algebraic closed form
    solution."""
    m = len(mu)
    ones = np.ones(m)

    SigmaInv = la.inv(Sigma)
    A = mu.T @ SigmaInv @ ones
    B = mu.T @ SigmaInv @ mu
    C = ones.T @ SigmaInv @ ones
    D = B * C - A * A
    l1 = (C * r - A) / D
    l2 = (B - A * r) / D
    return SigmaInv @ (l1 * mu + l2 * ones)


def minimum_variance_numeric(Sigma: np.ndarray, mu: np.ndarray, r: float) -> np.ndarray:
    """Calculate weights using Lagrangian multipliers and numeric solution (using
    scipy's root function)."""
    m = len(mu)
    ones = np.ones(m)

    L = Lagrangian(
        objective=lambda x: x.T @ Sigma @ x / 2,
        constraints=[
            lambda x: r - x.T @ mu,
            lambda x: 1 - x.T @ ones,
        ],
        x0=ones / m,
    )
    return L.fit()


def minimum_variance_numeric_slsqp(
    Sigma: np.ndarray, mu: np.ndarray, r: float
) -> np.ndarray:
    """Calculate weights using Lagrangian multipliers and numeric solution (using
    scipy's minimize function)."""
    from scipy.optimize import minimize

    m = len(mu)
    ones = np.ones(m)

    res = minimize(
        lambda x: x.T @ Sigma @ x / 2,
        ones / m,
        method="SLSQP",
        constraints=(
            {"type": "eq", "fun": lambda x: r - x.T @ mu, "jac": lambda x: -mu.values},
            {"type": "eq", "fun": lambda x: 1 - x.T @ ones, "jac": lambda x: -ones},
        ),
        tol=1e-15,
    )

    return res.x



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yabte',
 'yabte.backtest',
 'yabte.tests',
 'yabte.utilities',
 'yabte.utilities.plot',
 'yabte.utilities.plot.matplotlib',
 'yabte.utilities.plot.plotly',
 'yabte.utilities.portopt',
 'yabte.utilities.simulation']

package_data = \
{'': ['*'], 'yabte.tests': ['data/nasdaq/*']}

install_requires = \
['mypy>=1.8.0,<2.0.0',
 'pandas-stubs>=2.1.4.231227,<3.0.0.0',
 'pandas>=2.2.1,<3.0.0',
 'scipy>=1.10.0,<2.0.0']

setup_kwargs = {
    'name': 'yabte',
    'version': '0.4.1',
    'description': 'Yet another backtesting engine',
    'long_description': '# yabte - Yet Another BackTesting Engine\n\nPython module for backtesting trading strategies.\n\nFeatures\n\n* Event driven, ie `on_open`, `on_close`, etc. \n* Multiple assets.\n* OHLC Asset. Extendable (e.g support additional fields, e.g. Volatility, or entirely different fields, e.g. Barrels per day).\n* Multiple books.\n* Positional and Basket orders. Extendible (e.g. can support stop loss).\n* Batch runs (for optimization).\n* Captures book history including transactions & daily cash, MtM and total values.\n\nThe module provides basic statistics like book cash, mtm and total value. Currently, everything else needs to be deferred to a 3rd party module like `empyrical`.\n\n## Core dependencies\n\nThe core module uses pandas and scipy.\n\n## Installation\n\n```bash\npip install yatbe\n```\n\n## Usage\n\nBelow is an example usage (the economic performance of the example strategy won\'t be good).\n\n```python\nimport pandas as pd\n\nfrom yabte.backtest import Book, SimpleOrder, Strategy, StrategyRunner\nfrom yabte.tests._helpers import generate_nasdaq_dataset\nfrom yabte.utilities.plot.plotly.strategy_runner import plot_strategy_runner_result\nfrom yabte.utilities.strategy_helpers import crossover\n\n\nclass SMAXO(Strategy):\n    def init(self):\n        # enhance data with simple moving averages\n\n        p = self.params\n        days_short = p.get("days_short", 10)\n        days_long = p.get("days_long", 20)\n\n        close_sma_short = (\n            self.data.loc[:, (slice(None), "Close")]\n            .rolling(days_short)\n            .mean()\n            .rename({"Close": "CloseSMAShort"}, axis=1, level=1)\n        )\n        close_sma_long = (\n            self.data.loc[:, (slice(None), "Close")]\n            .rolling(days_long)\n            .mean()\n            .rename({"Close": "CloseSMALong"}, axis=1, level=1)\n        )\n        self.data = pd.concat(\n            [self.data, close_sma_short, close_sma_long], axis=1\n        ).sort_index(axis=1)\n\n    def on_close(self):\n        # create some orders\n\n        for symbol in ["GOOG", "MSFT"]:\n            df = self.data[symbol]\n            ix_2d = df.index[-2:]\n            data = df.loc[ix_2d, ("CloseSMAShort", "CloseSMALong")].dropna()\n            if len(data) == 2:\n                if crossover(data.CloseSMAShort, data.CloseSMALong):\n                    self.orders.append(SimpleOrder(asset_name=symbol, size=-100))\n                elif crossover(data.CloseSMALong, data.CloseSMAShort):\n                    self.orders.append(SimpleOrder(asset_name=symbol, size=100))\n\n\n# load some data\nassets, df_combined = generate_nasdaq_dataset()\n\n# create a book with 100000 cash\nbook = Book(name="Main", cash="100000")\n\n# run our strategy\nsr = StrategyRunner(\n    data=df_combined,\n    assets=assets,\n    strategies=[SMAXO()],\n    books=[book],\n)\nsrr = sr.run()\n\n# see the trades or book history\nth = srr.transaction_history\nbch = srr.book_history.loc[:, (slice(None), "cash")]\n\n# plot the trades against book value\nplot_strategy_runner_result(srr, sr)\n```\n\n![Output from code](https://raw.githubusercontent.com/bsdz/yabte/main/readme_image.png)\n\n## Examples\n\nJupyter notebook examples can be found under the [notebooks folder](https://github.com/bsdz/yabte/tree/main/notebooks).\n\n## Documentation\n\nDocumentation can be found on [Read the Docs](https://yabte.readthedocs.io/en/latest/).\n\n\n## Development\n\nBefore commit run following format commands in project folder:\n\n```bash\npoetry run black .\npoetry run isort . --profile black\npoetry run docformatter . --recursive --in-place --black --exclude _unittest_numpy_extensions.py\n```\n',
    'author': 'Blair Azzopardi',
    'author_email': 'blairuk@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bsdz/yabte',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.13',
}


setup(**setup_kwargs)



from pathlib import Path

import pandas as pd

from yabte.backtest.asset import OHLCAsset

data_dir = Path(__file__).parent / "data"
notebooks_dir = Path(__file__).parents[2] / "notebooks"


def generate_nasdaq_dataset():
    assets = []
    dfs = []
    for csv_pth in (data_dir / "nasdaq").glob("*.csv"):
        name = csv_pth.stem
        assets.append(OHLCAsset(name=name, denom="USD"))
        df = pd.read_csv(csv_pth, index_col=0, parse_dates=[0])
        df.columns = pd.MultiIndex.from_tuples([(name, f) for f in df.columns])
        dfs.append(df)

    return assets, pd.concat(dfs, axis=1)



from dataclasses import dataclass, field
from typing import Callable, List, Optional

import numpy as np
from scipy.optimize import OptimizeResult, root
from scipy.optimize._numdiff import approx_derivative


@dataclass(kw_only=True)
class Lagrangian:
    x0: np.ndarray
    objective: Callable[[np.ndarray], float]
    constraints: List[Callable[[np.ndarray], float]] = field(default_factory=list)
    optimize_result: Optional[OptimizeResult] = None

    def f(self, x):
        return self.objective(x)

    def g(self, x):
        return np.array([f(x) for f in self.constraints])

    def f_grad(self, x):
        return approx_derivative(self.f, x)

    def g_jac(self, x):
        return approx_derivative(self.g, x)

    def H(self, z):
        x, l = np.split(z, (self.x0.shape[0],))
        eq1 = self.f_grad(x)
        eq2 = []
        if self.constraints:
            eq1 += self.g_jac(x).T @ l
            eq2 = self.g(x)
        return np.array([*eq1, *eq2])

    def fit(self):
        n_constraints = len(self.constraints)
        res = root(self.H, x0=np.r_[self.x0, [1] * n_constraints])
        self.optimize_result = res
        return res.x[:-n_constraints]
